/*
 //////////////////////////////////////////////////////////////
 //  Program..: _DCJDHLP.PRG                                 //
 //  Author...:  Jack Duijf                                  //
 //  Date.....: Nov 30, 2004                                 //
 //  Notes....: eXPress++ Tooltip Processor (single thread)  //
 //                                                          //
 //  02-04-11  Help object per DCREAD (not per thread)       //
 //            This prevents memory leaks.                   //
 //                                                          //
 //////////////////////////////////////////////////////////////
 */

#include 'dcdialog.ch'
#include "Appevent.ch"
#include "dcmsg.ch"
#INCLUDE "thread.CH"

// Defines for specifying balloon tip type,
// see :ShowBalloonTip()
#ifndef XBP_TIPDEFAULT
#define XBP_TIPDEFAULT                  0
#define XBP_TIPINFO                     1
#define XBP_TIPWARNING                  2
#define XBP_TIPERROR                    3
#define XBP_TIPINFOLARGE                4
#define XBP_TIPWARNINGLARGE             5
#define XBP_TIPERRORLARGE               6
#endif

CLASS JD_MagicHelpThread FROM Thread

HIDDEN:

VAR      Terminated
VAR      aWindows
METHOD   __AddTooltip
METHOD   __TestToolTip
EXPORTED:

METHOD   Init()
METHOD   execute()
METHOD   terminate()
METHOD   CheckTermination()
METHOD   AddToolTip()
METHOD   RemoveToolTip()
ENDCLASS

* --------------

METHOD JD_MagicHelpThread:Init()
::Thread:init()                                          // Init Superclass Thread()
::AtStart      := {||ThreadObject():Execute()}
::terminated   := .F.
::aWindows     := {}
Return SELF

SYNC METHOD JD_MagicHelpThread:__AddTooltip(aData)
Aadd(::aWindows,aData)
Return

SYNC METHOD JD_MagicHelpThread:__TestToolTip()
LOCAL ni       := 0
LOCAL aDialog  := {}
LOCAL oXbp     := nil
local bError := ErrorBlock({|e|Break(e)})  //Zoli
BEGIN SEQUENCE            //Zoli
For ni := 1 to Len(::aWindows)                                          // All tooltip references
   aDialog     := ::aWindows[ni]
   If ValTyp(aDialog) = "A" .and. Len(aDialog) >= 2
      oXbp     := aDialog[1]
      If ValType(oXbp) = "O"
         If aDialog[2]
            oXbp:TestTooltip()
         Else
            oXbp:Destroy()
            aDialog[1]  := nil
         Endif
      Endif
   Endif
Next ni
ni             := 1
do while ni <= Len(::aWindows)
   aDialog     := ::aWindows[ni]
   if ValTyp(aDialog) = "A" .and. Len(aDialog) >= 2
      if ::aWindows[ni,1] = nil                                             // Just removed
         Aremove(::aWindows,ni)
      else
         ni++
      Endif
   Else
      Aremove(::aWindows,ni)
  endif
enddo
END SEQUENCE            //Zoli
ErrorBlock(berror)  //Zoli
Return
* --------------

METHOD JD_MagicHelpThread:Execute()

LOCAL ni
LOCAL bError      := {||Break()}
LOCAL aDialog     := {}
LOCAL oXbp        := nil
DO WHILE !::Terminated

   BEGIN SEQUENCE
      ::__TestToolTip()
    Sleep(10)
    ::Checktermination()
   recover  // Zoli
    ::Terminated := .t.
    exit
   END SEQUENCE    // Zoli moved here, because usually error occur at the sleep line
enddo
Return

METHOD JD_MagicHelpThread:terminate
RETURN ( ::terminated := .T. )

* ----------------

METHOD JD_MagicHelpThread:checkTermination
IF ::terminated
  ::terminated := .F.
  ::quit()
  Sleep(5)
ENDIF
RETURN self

* ---------------

SYNC METHOD JD_MagicHelpThread:AddToolTip(oTool)

LOCAL nThread     := ThreadId()
LOCAL nPos        := 0
LOCAL lAdded      := .F.
LOCAL i, lAddToolTip := .t., oToolTip

if ValType(oTool) = "O" .and. oTool:isDerivedFrom("JD_MagicHelp")
//
// 02-04-2011 J. Duijf
// Now for every DCREAD a tooltip object is added.
// Even for single thread apps, with multiple dialogs. This ensures
// that all objects are cleared after :RemoveTooltip(oHelp), and the CG
// can cleanup.
//
//   FOR i := 1 TO Len(::aWindows)
//     oToolTip := ::aWindows[i,1]
//     IF Valtype(oToolTip) == 'O' .AND. oToolTip:producerid == nThread
//       lAddToolTip := .f.
//       EXIT
//     ENDIF
//   NEXT
//   if lAddToolTip
      Aadd(::aWindows,{oTool,.T.})
      lAdded := .T.
//   endif
endif
Return lAdded

* -------------
/*
 J.Duijf 01-04-2011 Remove by oHelp reference
 */
SYNC METHOD JD_MagicHelpThread:RemoveToolTip(oHelp)
LOCAL nThread     := ThreadId()
LOCAL nPos        := 0
if ValType(oHelp) = "O"                                                       // Remove tooltip by reference
   nPos           := Ascan(::aWindows,{|a|Valtype(a[1])=='O' .AND. a[1] = oHelp})
else                                                                          // Remove tooltip for this thread
   //
   // The tooltip object should not be removed by threadId.
   // Left this code for just in case.
   //
   nPos           := Ascan(::aWindows,{|a|Valtype(a[1])=='O' .AND. ;
                                          IsMemberVar(a[1],'producerID') .AND. ;
                                          Valtype(a[1]:producerId)=='N' .AND. ;
                                          a[1]:producerid = nThread})
endif
if nPos > 0                                                                   // Help object found
    ::aWindows[nPos,2]   := .F.                                               // Mark for remove
    Sleep(1)
endif
Return

* ------------

CLASS JD_Magichelp

// Copied from _dcfunct.prg
//
// Modified to have a single Tooltip instance per window.
// Created and destroyed in DcRead
//

HIDDEN:
METHOD   DisplayToolTip()
METHOD   PaintTheTip()
VAR      oSaveXbp
VAR      xSavemp1
VAR      nSaveEvent
VAR      oLastXbp
VAR      lastmp1
VAR      lastmp2
VAR      nLastEvent
VAR      nLastMotionTime

// PROTECTED:
// VAR      nLastCellPos

PROTECTED:
VAR nEvent
VAR mp1
VAR mp2
VAR oXbp
VAR nToolTipTimeOut

EXPORTED:
VAR      producerID
VAR      oLastMotionXBP
VAR      oLastTipXBP
VAR      oBlockedXBP
VAR      nLastTipTime
VAR      nLastShowTime
VAR      aLastMotionPos
VAR      oTip
VAR      lTipIsShown
VAR      nTipSensitivity
VAR      aToolColor
VAR      cToolFont
VAR      nToolTime

METHOD   TestTooltip()
METHOD   ShowTip()
METHOD   HideTip()
METHOD   Init()
METHOD   Destroy()

ENDCLASS

METHOD JD_Magichelp:init()
IF DC_LangSet() == DCLANG_CHINESE
   ::nTipSensitivity := .5
ELSE
   ::nTipSensitivity := DC_ToolTipSensitivity()
ENDIF
::producerID         := ThreadID()
::aToolColor         := { GRA_CLR_BLACK, GRA_CLR_WHITE }
::nLastShowTime      := 0
::lTipIsShown        := .F.
::oLastMotionXBP     := NIL
::aLastMotionPos     := NIL
::nToolTipTimeOut    := DC_ToolTipTimeOut()
// ::nLastCellPos       := 0
::nLastMotionTime    := 0
RETURN SELF

* -----------------


METHOD JD_Magichelp:TestTooltip()
LOCAL bError,wolthiba:=.f.   //Zoli


//
// Check for events in the thread of the producer
//
// LOCAL nEvent, mp1:=0, mp2:=0, oXbp:=NIL, aThreadInfo

/*
aThreadInfo:= ThreadInfo(THREADINFO_TID      + ;
                         THREADINFO_SYSTHND  + ;
                         THREADINFO_FUNCINFO + ;
                         THREADINFO_TOBJ)

IF AScan(aThreadInfo,{|a|a[1]==::producerID}) == 0
  RETURN
ENDIF
*/

::nEvent := LastAppEvent(@::mp1,@::mp2,@::oXbp,::producerID)

IF Valtype(::oXbp) # 'O' ; return ; endif                                    // No XBP part, ignore event

IF(ValType(::oBlockedXBP)=="O" .AND. ::oXbp == ::oBlockedXBP ) ; return ; endif // Blocked item, ignore event  Zoli moved here before BEGIN SEQUENCE

bError := ErrorBlock({|e|Break(e)})  //Zoli
BEGIN SEQUENCE            //Zoli


if ::oXbp:isDerivedFrom('DC_SetTimerEvent')                                  // Ignore events from timer
   ::oXbp              := ::oSaveXbp                                         // Assume same event als last
   ::mp1               := ::xSavemp1
   ::nEvent            := ::nSaveEvent
ELSE
   ::oSaveXbp        := ::oXbp                                               // Save event parameters
   ::xSavemp1        := ::mp1
   ::nSaveEvent      := ::nEvent
ENDIF

RECOVER //Zoli
wolthiba:=.t. //Zoli

END SEQUENCE    // Zoli
ErrorBlock(bError)      //Zoli

if wolthiba    //Zoli
 RETURN        //Zoli
endif                //Zoli

//IF(oXbp:isDerivedFrom("XbpIWindow"))                                     // Not sure why,
//   oXbp := oXbp:setParent()
//ENDIF
* // Zoli moved bedore BEGIN SEQUENCE IF(ValType(::oBlockedXBP)=="O" .AND. ::oXbp == ::oBlockedXBP ) ; return ; endif // Blocked item, ignore event
IF Valtype(::oLastXbp) # 'O'                                               // First Event of this Dialog.
   ::oLastXbp        := ::oXbp
   ::nLastEvent      := ::nEvent
   ::lastmp1         := ::mp1
   ::lastmp2         := ::mp2                                                // Save event parameters
ELSEIF ::oXbp == ::oLastXbp     .AND. ;
   Valtype(::lastmp1) = 'A'   .AND. ;
   Valtype(::mp1) = 'A'         .AND. ;
   Valtype(::lastmp1[1]) == Valtype(::mp1[1]) .AND. ::lastmp1[1] == ::mp1[1] .AND. ;
   Valtype(::lastmp1[2]) == Valtype(::mp1[2]) .AND. ::lastmp1[2] == ::mp1[2] .AND. ;
   Valtype(::nLastEvent) == Valtype(::nEvent) .AND. ::nLastEvent == ::nEvent  // Same event as before, on same object
   IF ::nLastShowTime == -1; return ; endif                                // Already timed-out, ignore event
ELSEIF ::oXbp == ::oLastXbp     .AND. ;
   Valtype(::lastmp1) = 'A'   .AND. ;
   Valtype(::mp1) = 'A'         .AND. ;
    ((Valtype(::lastmp1[1])==Valtype(::mp1[1]) .AND. ::lastmp1[1] # ::mp1[1]) .OR. ;
     (Valtype(::lastmp1[2])==Valtype(::mp1[2]) .AND. ::lastmp1[2] # ::mp1[2])) // Event on same object as before, change of position
   ::nLastShowTime   := Seconds()                                          // Renew last show-time
   ::oLastXbp        := ::oXbp
   ::lastmp1         := ::mp1
   ::lastmp2         := ::mp2
   ::nLastEvent      := ::nEvent                                             // Save event parameters
   return
ENDIF
IF Valtype(::oLastTipXBP) == 'O' .AND. ;
   ::nLastShowTime >= 0          .AND. ;
   (Seconds() - ::nLastShowTime) > ::nToolTipTimeOut                       // Time-out tooltip expired
   ::nLastShowTime   := -1                                                 // Timed-out
   ::hideTip()
   ::oLastXbp        := ::oXbp
   ::lastmp1         := ::mp1
   ::lastmp2         := ::mp2
   ::nLastEvent      := ::nEvent                                   // Save event parameters
   return
ENDIF


bError := ErrorBlock({|e|Break(e)})  //Zoli
BEGIN SEQUENCE            //Zoli


IF(::nEvent == xbeM_Motion)  .AND. ;
   Valtype(::mp1) = 'A'      .AND. ;
   ::mp1[1] > 0 .AND. ::mp1[2] > 0                                             // A mouse position has been detected
   ::oBlockedXBP := NIL                                            // No blocked object
   IF ( ValType( ::oLastMotionXBP) == "O" .AND. ;
      ::oXbp == ::oLastMotionXBP )                                 // Same object as before, movement detected
      IF (::aLastMotionPos[1] == ::mp1[1]   .AND. ;
         ::aLastMotionPos[2] == ::mp1[2])                          // Mouse has NOT moved

         if (::nLastMotionTime=nil)       //Zoli
          ::nLastMotionTime:=Seconds()    //Zoli
         endif                            //Zoli


         IF ((Seconds() - ::nLastMotionTime) > ::nTipSensitivity)  // Sensivity time-out exeeded.
            // IF oXbp:isderivedFrom("XbpCellGroup")
            //   ::nLastCellPos := oXbp:cellFromPos(mp1)
            // ENDIF
            ::showTip(::oXbp)                                      // A tooltip is shown (if any)
         ENDIF
      ELSEIF ValType(::oLastTipXBP) == 'O' .AND. ;
         ::oXbp:setParent() == ::oLastTipXBP:setParent() .AND. ;
         (::oXbp:isderivedFrom("XbpCellGroup") .OR. ::oXbp:isDerivedFrom('XbpScrollBar'))    // Mouse has moved over a cell
         ::hideTip()
         ::oLastMotionXBP  := ::oXbp
         ::aLastMotionPos  := AClone(::mp1)
         ::nLastMotionTime := Seconds()
         IF( ::nLastTipTime>0 .AND. (Seconds()-::nLastTipTime)<=0.5 )
            // ::nLastCellPos := oXbp:cellFromPos(mp1)
            ::showTip(::oXbp)
         ENDIF
      ELSE
         ::aLastMotionPos := AClone(::mp1)
         ::nLastMotionTime := Seconds()
      ENDIF
   ELSEIF (ValType(::oLastTipXBP) == "O" .AND. ;
      ::oXbp:setParent() == ::oLastTipXBP:setParent() .AND.;
      ::oXbp != ::oLastTipXBP )
      ::hideTip()
      ::oLastMotionXBP  := ::oXbp
      ::aLastMotionPos  := AClone(::mp1)
      ::nLastMotionTime := Seconds()
      IF( ::nLastTipTime>0 .AND. (Seconds()-::nLastTipTime)<=0.5 )
         // IF oXbp:isderivedFrom("XbpCellGroup")
         //   ::nLastCellPos := oXbp:cellFromPos(mp1)
         // ENDIF
         ::showTip(::oXbp)
      ENDIF
   ELSE
      ::oLastMotionXBP  := ::oXbp
      ::aLastMotionPos  := AClone(::mp1)
      ::nLastMotionTime := Seconds()
      ::hideTip()
   ENDIF
ELSEIF ::nEvent != xbeP_Paint .AND. Valtype(::oXbp)='O'

 BEGIN SEQUENCE          //Zoli

   IF ::oXbp==::oLastTipXBP .AND. !::oXbp:isDerivedFrom("XbpCellGroup") .AND. !::oXbp:isDerivedFrom("XbpScrollBar")
      ::oBlockedXBP    := ::oXbp
   ENDIF

 END SEQUENCE            //Zoli


   ::oLastMotionXBP := NIL
   ::hideTip()
ENDIF

END SEQUENCE            //Zoli
ErrorBlock(bError)      //Zoli
RETURN

* ------------------

METHOD JD_Magichelp:showTip(oXbp)
LOCAL bError //Zoli
bError := ErrorBlock({|e|Break(e)})  //Zoli
BEGIN SEQUENCE                      //Zoli

  IF !::lTipIsShown
    ::DisplayToolTip(::oLastMotionXBP)
    IF Valtype(::oTip) = 'O'
      ::oTip:show()
      ::lTipIsShown  := .T.
      ::oLastTipXBP  := ::oLastMotionXBP
      ::nLastShowTime := Seconds()
    ELSEIF (!::oLastMotionXBP:isDerivedFrom("XbpCellGroup") .OR. !::oLastMotionXBP:isDerivedFrom("XbpScrollBar"))
      ::lTipIsShown  := .F.
      ::oLastTipXBP  := nil
    ENDIF
    ::oBlockedXBP := NIL
  ENDIF
END SEQUENCE            //Zoli
ErrorBlock(bError)      //Zoli
RETURN

* -------------------

METHOD JD_Magichelp:hideTip()
LOCAL bError  //Zoli
bError := ErrorBlock({|e|Break(e)})//Zoli
BEGIN SEQUENCE   //Zoli

  IF Valtype(::lTipIsShown) == 'L' .AND. ::lTipIsShown
*    bError := ErrorBlock({|e|Break(e)})  //Zoli
*    BEGIN SEQUENCE                       //Zoli
    ::oTip:hide()
    IF ::oTip:status() == XBP_STAT_CREATE
      ::oTip:destroy()
    ENDIF
    ::lTipIsShown := .F.
    ::nLastTipTime := Seconds()
*    END SEQUENCE             //Zoli
*    ErrorBlock(bError)       //Zoli
  ENDIF
END SEQUENCE   //Zoli
ErrorBlock(bError)  //Zoli
RETURN

* ------------------

METHOD JD_Magichelp:DisplayToolTip(oXbpRequestingHint)

LOCAL cText := "(nil)"
LOCAL nID   := 0
LOCAL aPos
LOCAL oHelpLink
LOCAL cFont, nColorFG, nColorBG, nDelayTime
LOCAL bError   := nil
IF oXbpRequestingHint:isDerivedFrom('XbpCrt')
  RETURN self
ENDIF
aPos  := DC_CalcAbsolutePosition(::aLastMotionPos,oXbpRequestingHint)
IF aPos[1] > 5000
  aPos[1] := 5000
ENDIF
IF aPos[2] > 5000
  aPos[2] := 5000
ENDIF
aPos[1] += 6
aPos[2] -= 16
oHelpLink := DC_GetHelpLink(oXbpRequestingHint,DCGUI_HELP_TOOLTIP)

IF (ValType(oHelpLink)=="O" .AND. oHelpLink:isDerivedFrom("MagicHelpLabel")) ;
  .OR. DC_ToolTipShowSource()

  IF Valtype(oHelpLink) == 'O'
    cText := oHelpLink:getToolTip()
  ELSE
    cText := ''
  ENDIF

  IF !Empty(cText)
    IF Valtype(cText) = 'A'
      ASize(cText,5)
      cFont := cText[2]
      nColorFG := cText[3]
      IF Valtype(nColorFG) == 'A'
        nColorFG := GraMakeRGBColor(nColorFG)
      ENDIF
      nColorBG := cText[4]
      IF Valtype(nColorBG) == 'A'
        nColorBG := GraMakeRGBColor(nColorBG)
      ENDIF
      nDelayTime := cText[5]
      cText := cText[1]
    ENDIF
    IF Valtype(cText) = 'B'
      BEGIN SEQUENCE
         bError   := ErrorBlock({||Break()})
      cText := DC_XtoC(Eval(cText,oXbpRequestingHint))
      RECOVER
         cText    := nil
      END SEQUENCE
      ErrorBlock(bError)
      IF cText == NIL
        RETURN self
      ENDIF
     elseif ValType(cText) = "N"
      cText := alltrim(str(cText))
     elseif ValType(cText) = "D"
      cText := dtoc(cText)
     elseif ValType(cText) = "L"
      cText := if(cText,"TRUE","FALSE")
    ENDIF
  ENDIF

  IF DC_ToolTipShowSource()
    IF !Empty(cText)
       cText += Chr(13)+Chr(10)+Chr(13)+Chr(10)
    ENDIF
    cText += DC_ShowSourceCode(oXbpRequestingHint)
  ENDIF

ENDIF

IF cText == '(nil)' .AND. oXbpRequestingHint:isDerivedFrom('XbpCellGroup') .AND. ;
   Valtype(oXbpRequestingHint:cargo) = 'A' .AND. Len(oXbpRequestingHint:cargo) >= 3 .AND. ;
   !Empty(oXbpRequestingHint:cargo[3])
  cText := oXbpRequestingHint:cargo[3]
ELSEIF cText == '(nil)' .AND. oXbpRequestingHint:isDerivedFrom('XbpScrollBar') .AND. ;
   Valtype(oXbpRequestingHint:cargo) = 'A' .AND. Len(oXbpRequestingHint:cargo) >= 3 .AND. ;
   !Empty(oXbpRequestingHint:cargo[3])
  cText := oXbpRequestingHint:cargo[3]
ENDIF

IF (Valtype(cText) $ 'C' .AND. ( cText # '(nil)' .AND. !Empty(cText) )) .OR. ;
    Valtype(cText) == 'O'
   IF !DC_ShowBalloonTip() .OR. ;
        !(IsMethod(oXbpRequestingHint,"showBalloonTip") .AND. ;
         oXbpRequestingHint:showBalloonTip(XBP_TIPINFO,,cText))
      ::oTip := XbpStatic():new()
      ::oTip:options := XBPSTATIC_TYPE_FGNDFRAME
      ::oTip:create(AppDesktop(),AppDesktop(), aPos, { 0 , 0 })
      IF Empty(cFont) .AND. !Empty(::cToolFont)
        cFont := ::cToolFont
      ENDIF
      IF Valtype(cFont) == 'C'
        ::oTip:setFontCompoundName(cFont)
      ELSEIF Valtype(cFont) == 'O'
        ::oTip:setFont(cFont)
      ENDIF
      ::oTip:cargo := { cText, cFont, nColorFG, nColorBG, nDelayTime }
      ::PaintTheTip(cText)
   ENDIF
ELSE
  ::oTip := nil
ENDIF
RETURN(SELF)

* ------------------

METHOD JD_Magichelp:PaintTheTip(cText)

LOCAL aAttr, oPS, nTextLen, aText, i, nOffset, aPoints, aSize, oBitMap, ;
      aOrigPos := ::oTip:currentPos(), aMaxSize, nMaxWidth, aPos

IF Valtype(cText) = 'O' .AND. cText:isDerivedFrom('XbpBitMap')

  oBitMap := cText
  aSize := {oBitMap:xSize,oBitMap:ySize }
  aPos := {aOrigPos[1]+15,aOrigPos[2]+10-aSize[2]}

  aMaxSize     := DC_DeskTopSize()

  nMaxWidth    := aMaxSize[1]
  IF aPos[1] + aSize[1] > aMaxSize[1]
    aPos[1] := aMaxSize[1] - aSize[1]
  ENDIF
  IF aPos[2] < 0
    aPos[2] := 0
  ENDIF

  ::oTip:setPos(aPos)
  ::oTip:setSize(aSize,.F.)

  oPS := ::oTip:lockPS()
  oBitMap:draw( oPS )

ELSE

  oPS := ::oTip:lockPS()

  cText := Strtran(cText,';',Chr(13)+Chr(10))
  cText := Strtran(cText,'<%059>',';')
  aText := DC_TokenArray(cText,Chr(13)+Chr(10))
  nTextLen := 0
  aSize := { 0,0 }
  FOR i := 1 TO Len(aText)
    aPoints := GraQueryTextBox( oPS, aText[i] )
    aSize[1] := Max(aSize[1],(aPoints[3,1] - aPoints[1,1]) + 8 ) // width
  NEXT

  IF aPoints == nil
    ::oTip:unLockPS( oPS)
    RETURN(SELF)
  ENDIF

  aSize[2] := ((aPoints[1,2] - aPoints[2,2]) * Len(aText)) + 4 // height

  ::oTip:setPos({::oTip:currentPos()[1], ;
                 ::oTip:currentPos()[2]-aSize[2]})

  ::oTip:unlockPS()

  IF ::oTip:currentPos()[1] + aSize[1] > AppDeskTop():currentSize()[1] .AND. ;
     ::oTip:currentPos()[1] < AppDeskTop():currentSize()[1]
     ::oTip:setPos( { AppDeskTop():currentSize()[1] - aSize[1], ;
                      ::oTip:currentPos()[2] - aSize[2] } )
  ENDIF

  IF ::oTip:currentPos()[2] < 0
    ::oTip:setPos( { ::oTip:currentPos()[1], 0 } )
  ENDIF

  ::oTip:setSize(aSize,.F.)
  oPS := ::oTip:lockPS()

  DEFAULT ::oTip:cargo[3] := ::aToolColor[1], ;
          ::oTip:cargo[4] := ::aToolColor[2]

  aAttr := Array( GRA_AA_COUNT )
  aAttr [ GRA_AA_COLOR ] := ::oTip:cargo[4]
  GraSetAttrArea( oPS, aAttr )
  GraBox( oPS, { 1, 1}, { aSize[ 1] + 4, aSize[ 2] + 4}, GRA_FILL)

  aAttr := Array( GRA_AS_COUNT )
  aAttr [ GRA_AS_COLOR ] := ::oTip:cargo[3]
  GraSetAttrString( oPS, aAttr )
  GraBox( oPS, {0,0}, {aSize[1] - 1, aSize[2] - 1})

  nOffset := 0
  FOR i := Len(aText) TO 1 STEP -1
    GraStringAt( oPS, {4,4+nOffset}, aText[i])
    nOffset += (aPoints[1,2] - aPoints[2,2])
  NEXT

ENDIF

::oTip:unLockPS( oPS)
RETURN(SELF)

* ----------------

METHOD JD_MagicHelp:Destroy()
::oSaveXbp        := nil
::xSavemp1        := nil
::nSaveEvent      := nil
::oLastXbp        := nil
::lastmp1         := nil
::lastmp2         := nil
::nLastEvent      := nil
::nLastMotionTime := nil
// ::nLastCellPos    := nil
::producerID      := nil
::oLastMotionXBP  := nil
::oLastTipXBP     := nil
::oBlockedXBP     := nil
::nLastTipTime    := nil
::nLastShowTime   := nil
::aLastMotionPos  := nil
::oTip            := nil
::lTipIsShown     := nil
::nTipSensitivity := nil
::aToolColor      := nil
::cToolFont       := nil
::nToolTime       := nil
::oXbp            := nil
Return

* ------------------

GETSETFUNCTION DC_ShowBalloonTip DEFAULT .f.

